#include <vector>
#include <cstdint>
#include <cstdlib>
#include <memory>
#include <iostream>

#include <taskflow_/taskflow.hpp>
#include <taskflow_/algorithm/for_each.hpp>


template <typename T>
struct Storage
{
    std::vector<std::size_t> shape{};
    T *ptr{nullptr};

    Storage(const std::vector<std::size_t> &shape)
        : shape{shape}
    {
        std::size_t numElements = 1;
        for (auto element : shape)
        {
            numElements *= element;
        }

        ptr = static_cast<T *>(std::aligned_alloc(sysconf(_SC_PAGESIZE), sizeof(T) * numElements));
    }

    ~Storage()
    {
        // std::cout << "storage freed";
        free(ptr);
    }

    void printStats()
    {
        std::cout << "Storage Stats: " << getSize() << "\n";
        std::cout << "dim: " << shape.size() << " : ";
        for (auto len : shape)
        {
            std::cout << len << ", ";
        }
        std::cout << std::endl;
    }

    std::int32_t getSize() const
    {
        std::int32_t size = 1;
        for (auto len : shape)
        {
            size *= len;
        }
        return size;
    }
};

template <typename T>
using Memory = std::shared_ptr<Storage<T>>;

template <typename T>
Memory<T> TranposeFromImageToTime_v3_block_tf_no_struct_one_taskflow(const Memory<T> in, std::size_t fastBlockSizeDim, std::size_t slowBlockSizeDim, tf::Executor &executor)
{
    std::cout << "bluib";
    // CDCS::Utility::ScopedConsoleMicrosecondTimer timer("transposing data block   fast_write   tf  one taskflow< " + std::to_string(fastBlockSizeDim) + " , " + std::to_string(slowBlockSizeDim) + " >");

    const auto dims = (*in).shape;
    const std::size_t X = dims[0];
    const std::size_t Y = dims[1];
    const std::size_t Z = dims[2];

    auto out = std::make_shared<Storage<T>>(std::vector<std::size_t>{Z, X, Y});

    const std::size_t imagesize = X * Y;

    const std::size_t fastDim = imagesize;
    const std::size_t slowDim = Z;

    T *in_ptr = in->ptr;
    T *out_ptr = out->ptr;

    // add regular patches
    std::size_t fastBlockPos = 0;
    std::size_t slowBlockPos = 0;

    tf::Taskflow taskflow;

    if (slowDim >= slowBlockSizeDim && fastDim >= fastBlockSizeDim)
    {
        while (slowBlockPos + slowBlockSizeDim <= slowDim)
        {

            fastBlockPos = 0;
            while (fastBlockPos + fastBlockSizeDim <= fastDim)
            {

                taskflow.emplace(
                    [in_ptr, out_ptr, fastBlockSizeDim, slowBlockSizeDim, fastBlockPos, slowBlockPos, fastDim, slowDim]()
                    {
                        for (std::size_t fast = 0; fast < fastBlockSizeDim; fast++)
                        {
                            for (std::size_t slow = 0; slow < slowBlockSizeDim; slow++)
                            {
                                out_ptr[slowBlockPos + fastBlockPos * slowDim + slow + fast * slowDim] = in_ptr[fastBlockPos + slowBlockPos * fastDim + fast + slow * fastDim];
                            }
                        }
                    });
                fastBlockPos += fastBlockSizeDim;
            }

            slowBlockPos += slowBlockSizeDim;
        }
    }

    std::size_t fastEnd = fastBlockPos;
    std::size_t slowEnd = slowBlockPos;

    std::size_t fastLeftover = fastDim - fastEnd;
    std::size_t slowLeftover = slowDim - slowEnd;

    // check for leftovers
    if (fastLeftover != 0 && slowDim >= slowBlockSizeDim)
    {
        slowBlockPos = 0;

        while (slowBlockPos + slowBlockSizeDim <= slowDim)
        {
            taskflow.emplace(
                [in_ptr, out_ptr, fastLeftover, slowBlockSizeDim, fastEnd, slowBlockPos, fastDim, slowDim]()
                {
                    for (std::size_t fast = 0; fast < fastLeftover; fast++)
                    {
                        for (std::size_t slow = 0; slow < slowBlockSizeDim; slow++)
                        {
                            out_ptr[slowBlockPos + fastEnd * slowDim + slow + fast * slowDim] = in_ptr[fastEnd + slowBlockPos * fastDim + fast + slow * fastDim];
                        }
                    }
                });
            slowBlockPos += slowBlockSizeDim;
        }

        slowBlockPos += slowBlockSizeDim;
    }

    // check for leftovers
    if (slowLeftover != 0 && fastDim >= fastBlockSizeDim)
    {
        fastBlockPos = 0;

        while (fastBlockPos + fastBlockSizeDim < fastDim)
        {
            taskflow.emplace(
                [in_ptr, out_ptr, fastBlockSizeDim, slowLeftover, fastBlockPos, slowEnd, fastDim, slowDim]()
                {
                    for (std::size_t fast = 0; fast < fastBlockSizeDim; fast++)
                    {
                        for (std::size_t slow = 0; slow < slowLeftover; slow++)
                        {
                            out_ptr[slowEnd + fastBlockPos * slowDim + slow + fast * slowDim] = in_ptr[fastBlockPos + slowEnd * fastDim + fast + slow * fastDim];
                        }
                    }
                });
            fastBlockPos += fastBlockSizeDim;
        }

        fastBlockPos += fastBlockSizeDim;
    }

    if (slowLeftover != 0 && fastLeftover != 0)
    {

        taskflow.emplace(
            [in_ptr, out_ptr, fastLeftover, slowLeftover, fastEnd, slowEnd, fastDim, slowDim]()
            {
                for (std::size_t fast = 0; fast < fastLeftover; fast++)
                {
                    for (std::size_t slow = 0; slow < slowLeftover; slow++)
                    {
                        out_ptr[slowEnd + fastEnd * slowDim + slow + fast * slowDim] = in_ptr[fastEnd + slowEnd * fastDim + fast + slow * fastDim];
                    }
                }
            });
    }
    executor.run(taskflow).wait();
    return out;
}
